// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
#include <string.h>

#include <mgrtree/Assert.h>
#include <mgrtree/System.h>
#include <mgrtree/Rtree.h>
#include "SegIncid.h"
#include "T_xx.h"

#undef EndOfIncludesSection

/*------------------------------------------------------------------------- o
  private definitions
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
HsoSingleSignature::get (
Abscissa & baseAb,
Abscissa & endsAb)
{
  XyVectorI  u (baseAb);
  for (short o0 = -1; o0 <= 1; o0 += 2) {
    XyVectorI  v (baseAb.start(), endsAb.end(o0));
    hso (o0) = u.crossProductSign (v);
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
HalfspaceOrientSignature::get (
Abscissa & priAb,
Abscissa & secAb)
{
  pri.get (priAb, secAb);
  sec.get (secAb, priAb);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
short
SegWid::order (
SegWid &        sec,
AbAbSignature & abAbSignature)
{
  short abAbOrder = abAbSignature.order ();
  if ( abAbOrder != 0 )
    return abAbOrder;
  return  id.order (sec.id);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegLocSpotter::init (
SegWid & baseSeg)
{
  this->baseSeg = baseSeg;
  locator.setRange (baseSeg.range());
  locator.setPreStart ();
  singleStatus.init();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegIncidenceSpot::setPair (
SegLocSpotter & spotter)
{
  priSeg = spotter.priSeg();
  secSeg = spotter.secSeg();
  pairSignature = spotter.pairSignature;
  singleStatus.setInvalid();
  basicType = pairSignature.isInnerCrossing() ? innerCrossing() :
              pairSignature.isPartialCoalescence() ? partialCoalescence() :
              pairSignature.isDuplicity() ? duplicity() : -1;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegIncidenceSpot::setSingle (
SegLocSpotter & spotter)
{
  priSeg = spotter.priSeg();
  secSeg = priSeg; // dummy
  pairSignature.setInvalid();
  singleStatus = spotter.singleStatus;
  basicType = freeEnd();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
short
SegIncidenceSpot::order (
SegIncidenceSpot & spot)
{
  // first, look after the type
  if ( basicType < spot.basicType )
    return  -1;
  if ( basicType > spot.basicType )
    return  1;

  short priOrder = priSeg.ab.order (spot.priSeg.ab);
  if ( priOrder != 0 )
    return  priOrder;

  return  secSeg.ab.order (spot.secSeg.ab);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
CsizeString<128>
SegIncidenceSpot::cstr ()
{
  CsizeString<128>   s;
  char               s0[128];
  if ( isInnerCrossing() ) {
    char               hs[128];
    sprintf (hs, "%d,%d,%d,%d", pairSignature.hs.hso(-1,-1),
                                pairSignature.hs.hso(-1,1),
                                pairSignature.hs.hso(1,-1),
                                pairSignature.hs.hso(1,1));
    sprintf (s0, "Inner crossing: segments %s and %s, hs: (%s)",
                 priSeg.cstr().s(), secSeg.cstr().s(), hs);
  }
  else if ( isPartialCoalescence() ) {
    char               lx[128];
    sprintf (lx, "%d,%d,%d,%d", pairSignature.lx.lo(-1,-1),
                                pairSignature.lx.lo(-1,1),
                                pairSignature.lx.lo(1,-1),
                                pairSignature.lx.lo(1,1));
    sprintf (s0, "Partial coalescence: segments %s and %s, lx: (%s)",
                 priSeg.cstr().s(), secSeg.cstr().s(), lx);
  }
  else if ( isDuplicity() ) {
    sprintf (s0, "Duplicity: segment %s", priSeg.cstr().s());
  }
  else if ( isFreeEnd() ) {
    sprintf (s0, "Free end: segment %s, (%s)", priSeg.cstr().s(),
                 singleStatus.isNoEndCovered() ? "both ends" :
                 singleStatus.end (priSeg.o0()).isCovered ? "second end" :
                                                            "first end");
  }
  else {
    sprintf (s0, "???");
  }
  s = s0;
  return s;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegLocSpotter::processPair ()
{
  if ( priSeg().isEqual (secSeg()) ) {
    return  SUCCESS;
  }
  pairSignature.get (priSeg().ab, secSeg().ab);

  if ( !pairSignature.isIncident() )
    return  SUCCESS;
  for (short o = -1; o <= 1; o += 2 ) {
    if ( pairSignature.isEndCovered (o) )
      endStatus(o).isCovered = TRUE;
  }
  if ( pairSignature.isRegularIncident() ) {
    return  SUCCESS;
  }
  // irregular incidence
  if ( priSeg().order (secSeg(), pairSignature) != -1 )
    return  SUCCESS;  // wait until the pair appears in reverse order
  SegIncidenceSpot spot;
  spot.setPair (*this);
  return
  spotList->addMember (spot);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegLocSpotter::processSingle ()
{
  if ( singleStatus.isRegular() )
    return  SUCCESS;
  SegIncidenceSpot spot;
  spot.setSingle (*this);
  return
  spotList->addMember (spot);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegLocSpotter::processAll ()
{
  while ( SUCCESS == nextStep() ) {
    processPair();
  }
  processSingle();
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegIncidSpotList::print ()
{

  if ( isEmpty() ) {
    printf ("\nNo singularity found in segment incidences.\n");
    return;
  }
  printf ("\nSingularities found in segment incidences (%d):\n", n);
  ListPos <SegIncidenceSpot>  pos (this);
  while ( SUCCESS == pos.nextStep() ) {
    printf ("%s\n", pos.memb->cstr().s());
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegIncidSpotter::processAll ()
{
  SegLocator    locator_base (this->segSet->rtree);
  SegLocSpotter locSpotter (&spotList, this->segSet->rtree);

  XyRange    outerRange;
  outerRange.setUnlimited ();

  locator_base.setRange (outerRange);

  spotList.removeAll();

  MgSystem::startTime ();
  printf ("Processing segment incidences...\n");

  while ( SUCCESS == locator_base.nextStep() ) {     // outer loop
    locSpotter.init (locator_base.seg);
    locSpotter.processAll ();
  }
  double elapsedTime = MgSystem::difTime();
  printf ("Segment incidences processed in %.2f seconds\n",
           elapsedTime);

  if ( spotList.n <= 200 ) // make the results to look finer
    spotList.bubblesort();

  spotList.print();

  printf ("\nFocused time in seconds: %.2f\n", elapsedTime);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegIncidSpotter::processAll (
const char * fileName)
{
  SegRtreeSet      segs;
  SegIncidSpotter  spotter (&segs);
  segs.createFromTextFile (fileName);
  spotter.processAll ();
}

/*------------------------------------------------------------------------- o
o ------------------------------------------------------------------------*/
static void
test0 ()
{
  SegRtreeSet      segs;
  SegIncidSpotter  spotter (&segs);
  segs.createFromTextFile ("e:\\w\\1\\mgrtree\\out\\SegList0.txt");
//  segs.createFromTextFile ("e:\\w\\2\\mgrtlib\\testdata\\SegListE.txt");
  spotter.processAll ();
}

/*------------------------------------------------------------------------- o
o ------------------------------------------------------------------------*/
void
testSegIncidSpotter ()
{
  test0 ();
}



