// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

#ifndef  __segincidH__
#define  __segincidH__

#include <mgrtree/BasDef.h>
#include <mgrtree/CsizeStr.h>
#include "SegRtSet.h"

#undef EndOfIncludesSection

// 2d vector with integer coordinates
// assumption: all numeric operation are exact provided its operands
//             are sufficiently small
class XyVectorI {
 public:
  int     x;
  int     y;
 public:
  XyVectorI () {};
  XyVectorI (AbEndPoint & start, AbEndPoint & end) {
    fromStartEnd (start, end); }
  XyVectorI (Abscissa & ab) { fromAbscissa (ab); }
 public:
  void fromStartEnd (AbEndPoint & start, AbEndPoint & end) {
    x = end.x - start.x; y = end.y - start.y; }
  void fromAbscissa (Abscissa & ab) {
    fromStartEnd (ab.start(), ab.end()); }
 public:
  short crossProductSign (XyVectorI & v) {
    __int64 left =  (__int64) x * v.y;
    __int64 right = (__int64) y * v.x;
    return left < right ? -1 : left > right ? 1 : 0; }
};

// Given two abscissae, the first one forming a halfspace partition, the
// second one supplying its ends.
// The halfspace orientation of -1/+1 end E is the sign of
// the halfspace induced by the supporting line of the first abscissa
// having E in its interior (0 means E is on the supporting line).
class HsoSingleSignature {
 public:
  short    os[2];   // orientations
 public:
  void init () {}
 public:
  short & hso (short o0) { return o0 == -1 ? os[0] : os[1]; };
 public:
  void setInvalid () { hso(-1) = hso(1) = -2; }
 public:
  void get (Abscissa & baseAb, Abscissa & endsAb);
 public:
  sbool isCollinear () { return hso(-1) == 0 && hso(1) == 0; }
  sbool isCrossing () { return !isCollinear() && hso(-1) != hso(1); }
  sbool isInnerCrossing () { return isCrossing() && hso(-1) == -hso(1); }
 public:
  // returns -1 iff base abscissa is crossed by ends abscissa from the -1
  // halfspace
  short crossSign () { return MG_sign (hso(-1) - hso(1)); }
};

// Given two abscissae, each forming a halfspace partition.
// The halfspace orientation of -1/+1 end E of an abscissa is the sign of
// the halfspace induced by the supporting line of the opposite abscissa
// having E in its interior (0 means E is on the supporting line).
class HalfspaceOrientSignature {
 public:
  HsoSingleSignature pri; // halfspace induced by the primary abscissa
  HsoSingleSignature sec; //                          secondary
 public:
  void init () {}
  void setInvalid () { pri.setInvalid(); sec.setInvalid(); }
 public:
  HsoSingleSignature & prisec (short o) { return o == -1 ? pri : sec; }
  short & hso (short o, short o0) { return prisec(o).hso(o0); };
 public:
  void get (Abscissa & priAb, Abscissa & secAb);
 public:
  sbool isCollinear () { return pri.isCollinear(); }
  sbool isCrossing () { return pri.isCrossing() && sec.isCrossing(); }
  sbool isRegularCrossing () { return isCrossing() &&
    !pri.isInnerCrossing() && !sec.isInnerCrossing(); }
 public:
  // works if isCrossing()
  sbool isEndCovered (short o0) { return sec.hso (o0) == 0; }
 public:
  // slope order of abscissae; works if isCrossing()
  short slopeOrd () { return  pri.crossSign(); }
};

// Given an abscissa and a point P. The -1/+1 part of the signature
// indicates the lexicographical order orientation of P related to -1/+1
// end E of the abssissa. (-1 means P before (left), 0 means P equal.)
class LexOrdSingleSignature {
 public:
  short    os[2];   // orientations
 public:
  short & lo (short o0) { return o0 == -1 ? os[0] : os[1]; };
 public:
  void setInvalid () { lo(-1) = lo(1) = -2; }
 public:
  void get (Abscissa & ab, AbEndPoint & p) {
   for (short o = -1; o <= 1; o += 2 ) lo (o) = p.lexOrd (ab.end(o)); }
 public:
  sbool isBetween () { return lo(-1) != -1 && lo(1) != 1; }
  sbool isInnerBetween () {
    return isBetween() && lo(-1) != 0 && lo(1) != 0; }
};

// Given two abscissae, the first one as a base.
// The signature indicates the lex order orientation of the second
// abscissa's ends to the base abscissa's ends.
class LexOrderSignature {
 private:
  LexOrdSingleSignature  loss[2]; // lex order single signatures
 public:
  void init () {}
  void setInvalid () { los(-1).setInvalid(); los(1).setInvalid(); }
 public:
  LexOrdSingleSignature & los (short o) {
                                    return o == -1 ? loss[0] : loss[1]; }
  short & lo (short o, short o0) { return los(o).lo(o0); }
 public:
  void get (Abscissa & base, Abscissa & sec) {
   for (short o = -1; o <= 1; o += 2 ) los(o).get (base, sec.end(o)); }
 public:
  // lex order intervals overlapping
  sbool isOverlapping () {
    return los(-1).isBetween() || los(1).isBetween() ||
           lo(-1,-1) != lo(1,1); }  // 2nd args are unsignificant
  sbool isRegularOverlapping () {
    return isOverlapping() && (lo(-1,1) == 0 || lo(1,-1) == 0); }
//  sbool isRegularOverlapping (short o) {
//    return isOverlapping() && lo(-o,o) == 0; }
 public:
  sbool isEndCovered (short o) {
    return  (lo(-o,o) == 0 || lo(-o,o) == -o) && lo(o,o) == o; }
 public:
  short intervalOrd () { return lo(-1,-1) != 0 ? -lo(-1,-1) : -lo(1,1); }
  sbool isDuplicity () { return intervalOrd() == 0; }
};

// classifying relations between two abscissae
class AbAbSignature {
 public:
  HalfspaceOrientSignature   hs;
  LexOrderSignature          lx;  // significant in collinearity case,
                                  // used directly for the duplicity check
 public:
  void setInvalid () { hs.setInvalid(); lx.setInvalid(); }
 public:
  void get (Abscissa & pri, Abscissa & sec) {
    hs.get (pri, sec); lx.get (pri, sec); }
 public:
  sbool isIncident () { return hs.isCrossing() ||
                               hs.isCollinear() && lx.isOverlapping(); }
  sbool isRegularIncident() {
    return hs.isCrossing() ? hs.isRegularCrossing() :
           hs.isCollinear() ? lx.isRegularOverlapping() : FALSE; }
 public:
  // 3 basic categories of relations between two irregularly incident
  // abscissae
  sbool isInnerCrossing () {
    return hs.isCrossing() && !hs.isRegularCrossing(); }
  sbool isPartialCoalescence() {
    return hs.isCollinear() && !lx.isRegularOverlapping() &&
                               !lx.isDuplicity(); }
  sbool isDuplicity() { return lx.isDuplicity(); }
 public:
  sbool isEndCovered (short o0) {
    return  hs.isCrossing() ? hs.isEndCovered (o0) :
                              lx.isEndCovered (o0); }
 public:
   short order () { short slopeOrd = hs.slopeOrd();
     if ( slopeOrd != 0 ) return slopeOrd;
     else                 return lx.intervalOrd();
   }
};


class EndStatus {
 public:
  sbool   isCovered;
 public:
  void init () { isCovered = FALSE; }
  EndStatus () { init(); }
 public:
  void setInvalid () { isCovered = -1; }
 public:
  sbool isRegular () { return isCovered; }
};

class SingleStatus {
 public:
  EndStatus  ends[2];
 public:
  EndStatus & end (short o) { return o == -1 ? ends[0] : ends[1]; }
 public:
  void setInvalid () { end(-1).setInvalid(); end(1).setInvalid(); }
  void init () { end(-1).init(); end(1).init(); }
 public:
  sbool isNoEndCovered () {
    return !end(-1).isCovered && !end(1).isCovered; }
  sbool isRegular () { return end(-1).isRegular() && end(1).isRegular(); }
};

class SegIncidSpotList;
class SegLocSpotter;

class SegIncidenceSpot {
 public:
  static short innerCrossing () { return 1; }
  static short partialCoalescence () { return 2; }
  static short duplicity () { return 3; }
  static short freeEnd () { return 4; }
 public:
  sbool isInnerCrossing () { return basicType == innerCrossing(); }
  sbool isPartialCoalescence () {return basicType == partialCoalescence();}
  sbool isDuplicity () { return basicType == duplicity(); }
  sbool isFreeEnd () { return basicType == freeEnd(); }
 public:
  short         basicType;
  SegWid        priSeg;
  SegWid        secSeg;        // unsignificant in case of freeEnd
  AbAbSignature pairSignature; // wasteful in its size
  SingleStatus  singleStatus;  //        dtto
 public:
  SegIncidenceSpot * next;
 public:
  void init () { basicType = -1; }
  SegIncidenceSpot () { init(); }
 public:
  void setPair (SegLocSpotter &);
  void setSingle (SegLocSpotter &);  // free end
 public:
  short order (SegIncidenceSpot & spot);
 public:
  CsizeString<128> cstr();
};


// segment local spotter
class SegLocSpotter {
 private:
  SegIncidSpotList * spotList; // resulting list
 public:
  SegWid             baseSeg;  // seg to spot singularities for
  SegLocator         locator;  // raw specification of incident segs
  SingleStatus       singleStatus; // for free end detection
  AbAbSignature      pairSignature;
 public:
  EndStatus & endStatus (short o) { return singleStatus.end(o); }
 public:
  void init () { spotList = NULL; }
  void init (SegIncidSpotList * spotList, SegRtree & rtree) {
    init(); this->spotList = spotList; locator.init (rtree); }
  SegLocSpotter (SegIncidSpotList * spotList, SegRtree & rtree) {
    init (spotList, rtree); }
  void init (SegWid & baseSeg);
 public:
  SegWid & priSeg () { return baseSeg; }
  SegWid & secSeg () { return locator.seg; }
 public:
  int nextStep () { return locator.nextStep(); }
  int processPair ();
  int processSingle (); // pastEndProcess
  int processAll ();
};

class SegIncidSpotList : public List <SegIncidenceSpot> {
 public:
  void print ();
};

class SegIncidSpotter {
 private:
  SegRtreeSet *    segSet;
 public:
  SegIncidSpotList spotList;
 public:
  void init () { segSet = NULL; }
  SegIncidSpotter () { init(); }
  SegIncidSpotter (SegRtreeSet * segSet) { init(); this->segSet = segSet; }
 public:
  void processAll ();
 public:
  static void processAll (const char * fileName);
};

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/

#endif      /* __segincidH__ */






