// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
#include <string.h>
#include <malloc.h>

#include <mgrtree/Assert.h>
#include "SegList.h"
#include "T_xx.h"

#undef EndOfIncludesSection

/*------------------------------------------------------------------------- o
  private definitions
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
CsizeString<64>
Abscissa::cstr ()
{
  char             s[128];
  sprintf (s, "(%d,%d),(%d,%d)", start().x, start().y, end().x, end().y);
  return  CsizeString<64> (s);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
CsizeString<64>
Segment::cstr ()
{
  char             s[128];
  short            o = o0();
  sprintf (s, "(%d,%d),(%d,%d)", ab.end( o).x, ab.end( o).y,
                                 ab.end(-o).x, ab.end(-o).y);
  return  CsizeString<64> (s);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Segment::fromString (
const char * s)
{
  AbEndPoint  p0, p1;
  short       color = 0;
  int         numFieldsRead;
  numFieldsRead =
  sscanf (s, "(%d,%d),(%d,%d)%s", &p0.x, &p0.y, &p1.x, &p1.y, &color);

  if ( numFieldsRead < 4 )
    return  FAILURE;

  fromPts (p0, p1);
  f.color = color;
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegList::addSegment (
Segment & seg)
{
  if ( SUCCESS != add (SegEntry(seg)) )
    return  FAILURE;
  if ( seg.isGray() )
    graySegsBuf.add (SegId (lastMembP()));
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegList::print ()
{
  if ( isEmpty() ) {
    printf ("\nNo entries in the segment list.\n");
    return  SUCCESS;
  }
  SegListPos pos (this);
  printf ("\nSegments in list (# of %d):\n", n);
  while ( SUCCESS == pos.nextStep() ) {
    CsizeString<64> s = pos.memb->cstr();
    printf ("%s\n", s.s());
  }
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
char *
SegListReader::strContentPastEnd (
char * s)
{
  // start of white space before comment(#) or line end(\n) whichever
  // comes first
  char * commentStart = strchr (s, '#');
  char * lineEnd = strchr (s, '\n');
  int len = strlen (s);
  char * pastEnd = commentStart == NULL ? lineEnd :
                   lineEnd == NULL ? commentStart :
                   commentStart < lineEnd ? commentStart : lineEnd;
  if ( pastEnd == NULL )
    pastEnd = (char *) s + len;
  while ( pastEnd > s && pastEnd[-1] == ' ') {
    pastEnd--;
  }
  return  pastEnd;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
sbool
SegListReader::isAsEmptyString (
char * s)
{
  return  s == strContentPastEnd (s);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegListReader::init ()
{
  segList = NULL;
  fileP = NULL;
  strcpy (fileName, "");
  curLineIdx = -1;
  strcpy (curLine, "");
  numPointSegments = 0;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegListReader::openFile (
const char * fileName)
{
  assert (!isFileOpened());
  fileP = fopen (fileName, "rt");
  if ( fileP == NULL )
    return FAILURE;
  assert (strlen (fileName) < sizeof (this->fileName));
  strcpy (this->fileName, fileName);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegListReader::closeFile ()
{
  assert (isFileOpened());
  fclose (fileP);
  fileP = NULL;
  strcpy (fileName, "");
  curLineIdx = -1;
  strcpy (curLine, "");
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegListReader::nextLineStep_read ()
{
  int  width = sizeof (curLine);
  char formatStr[64];

  // everything to '\n', possibly restricted to width
  sprintf (formatStr, "%%%d[^\n]\n", width);
  int numFieldsRead =
  fscanf (fileP, formatStr, curLine);
  curLineIdx++;
  return  numFieldsRead == 1 ? SUCCESS : FAILURE;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
SegListReader::checkSegmentSize (
sbool *    isAccepted,
Segment &  seg)
{
  if ( seg.ab.isPoint() ) {
    printf ("Warning: segment %s ignored (a point segment)\n",
             seg.cstr().s());
    numPointSegments++;
    *isAccepted = FALSE;
  }
  else {
    *isAccepted = TRUE;
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegListReader::nextLineStep ()
{
  if ( SUCCESS != nextLineStep_read () )
    return  FAILURE;

  // truncate read line to it's content only
  char * sP;
  if ( NULL != (sP = strContentPastEnd (curLine)) )
    *sP = '\0';

  if ( isAsEmptyString (curLine) )  // empty line, skip it
    return  SUCCESS;

  Segment  seg;
  if ( SUCCESS != seg.fromString (curLine) ) {
    printf ("\n%s(%d): string \"%s\" not recognized!\n", fileName,
            curLineIdx + 1, curLine);
    return  SUCCESS;
  }
  sbool  isSegAccepted;
  checkSegmentSize (&isSegAccepted, seg);
  if ( !isSegAccepted )
    return  SUCCESS;
  if ( segList == NULL ) { // means diagnostics
    printf ("seg: %s\n", seg.cstr().s());
    return  SUCCESS;
  }
  segList->addSegment (seg);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegListReader::readAll ()
{
  printf ("\nReading segments from file %s...\n", fileName);
  while ( SUCCESS == nextLineStep() ) {
//    printf ("%d: %s\n", curLineIdx, curLine);
  }
  printf ("%d segments in segment list\n",
          segList == NULL ? 0 : segList->n);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegListReader::readFile (
const char * fileName)
{
  if ( SUCCESS != openFile (fileName) ) {
    printf ("\nCannot open file %s\n", fileName);
    return  FAILURE;
  }
  readAll ();
  closeFile ();
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SegList::createFromTextFile (
const char * fileName)
{
  assert (isEmpty());
  return  SegListReader (this).readFile (fileName);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
static void
testSegListReader ()
{
  SegList        segList;
  SegListReader  reader (&segList);
  reader.readFile ("e:\\w\\1\\mgrtree\\segincid\\in\\test1.txt");
  segList.print();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
testSegList ()
{
  testSegListReader ();
}



