// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

#ifndef  __seglistH__
#define  __seglistH__

#include <stdio.h>

#include <mgrtree/BasDef.h>
#include <mgrtree/CsizeStr.h>
#include "../include/ArList.h"

#undef EndOfIncludesSection

class AbEndPoint {
 public:
  int     x;
  int     y;
 public:
  AbEndPoint () {};
  AbEndPoint (int x, int y) : x(x), y(y) {}
 public:
  short lexOrd (AbEndPoint & p) {
    return x < p.x ? -1 : x > p.x ? 1 : y < p.y ? -1 : y > p.y ? 1 : 0; }
  sbool isEqual (AbEndPoint & p) { return x == p.x && y == p.y; }
};

class XyRange;

// abscissa with lexicographically ordered ends with integer coordinates
class Abscissa {
 private:
  AbEndPoint  ends[2];
 public:
  Abscissa () {}
  Abscissa (AbEndPoint & p0, AbEndPoint & p1, short * o0P = NULL) {
    fromPts (p0, p1, o0P); }
 public:
  void fromPts (AbEndPoint & p0, AbEndPoint & p1, short * o0P = NULL) {
    short  o0 = p0.lexOrd (p1);
    if ( o0 == -1 ) { ends[0] = p0; ends[1] = p1; }
    else            { ends[0] = p1; ends[1] = p0; }
    if ( o0P != NULL ) *o0P = o0;
  }
 public:
  AbEndPoint & end (short o) { return o == -1 ? ends[0] : ends[1]; }
  AbEndPoint & start () { return  ends[0]; }
  AbEndPoint & end () { return  ends[1]; }
  XyRange range ();
 public:
  sbool isPoint () { return start().isEqual (end()); }
 public:
  short order (Abscissa & ab) {
    short startsOrder = start().lexOrd (ab.start());
    return  startsOrder != 0 ? startsOrder : end().lexOrd (ab.end());
  }
 public:
  CsizeString<64> cstr();
};

class Segment {
 public:
  Abscissa    ab;
  struct {
    Ushort    o0 : 1;  // inborn ab orientation, 1 for reversed
    Ushort    color : 3; // segment's color
    Ushort    unused : 12;
  }  f;  // flags
  short       inNodeIdx;  // segment's entry in-node index
                          // placed here (and not in SegEntry) due to
                          // the alignment effect
 public:
  void init () { f.color = 0; f.unused = 0; inNodeIdx = -1; }
  Segment () { init(); }
  void fromPts (AbEndPoint & p0, AbEndPoint & p1) { short o0;
    init (); ab.fromPts (p0, p1, &o0);  f.o0 = o0 == 1; }
  Segment (AbEndPoint & p0, AbEndPoint & p1) { fromPts (p0, p1); }
 public:
  XyRange range ();
  short o0 () { return  f.o0 == 0 ? -1 : 1; }
 public:
  int fromString (const char * s);
  CsizeString<64> cstr();
 public:
  sbool isGreen ()  { return f.color == 0; }
  sbool isRed ()    { return f.color == 1; }
  sbool isYellow () { return f.color == 2; }
  sbool isViolet () { return f.color == 3; }
  sbool isGray ()   { return f.color == 4; }
};

class SegEntry : public Segment {
 public:
  SegEntry () {}
  SegEntry (Segment & seg) : Segment (seg) {}
 public:
  Segment & segment() { return *this; }
  void setSegment (Segment & seg) { segment() = seg;}
 public:
  short order (SegEntry & entry) {
    return  ArListPos<SegEntry>(this).order (ArListPos<SegEntry>(&entry));}
};

union RecId;

class SegId {
 public:
  SegEntry * entry;
 public:
  void init () { entry = NULL; }
  SegId () { init(); }
  SegId (SegEntry * entry) : entry (entry) {}
  SegId (RecId & recId);
 public:
  void setNull () { entry = NULL; }
  sbool isNull () { return entry == NULL; }
  sbool isEqual (SegId & id) { return entry == id.entry; }
 public:
  short order (SegId & id) {
    return isNull() ? (id.isNull() ? 0 : -1) :
           id.isNull() ? 1 : entry->order (*id.entry); }
};

class AbAbSignature;

// segment with its id
class SegWid : public Segment {
 public:
  SegId  id;
 public:
  SegWid () {}
  SegWid (Segment & seg) : Segment (seg) { id.setNull(); }
 public:
  Segment & segment() { return *this; }
  void set (Segment & seg) { segment() = seg; id.setNull(); }
  void set (Segment & seg, SegId & id) { segment() = seg; this->id = id; }
 public:
  sbool isEqual (SegWid & seg) { return id.isEqual (seg.id); }
  short idOrder (SegWid & seg) { return id.order (seg.id); }
  short order (SegWid &, AbAbSignature &);
};

class GraySegsBuf {
 public:
  short     n;
  SegId     ids[10];
 public:
  void init () { n = 0; }
  GraySegsBuf () { init(); }
 public:
  short maxN () { return sizeof (ids) / sizeof (ids[0]); }
  int add (SegId & id) { if ( n >= maxN() ) return  FAILURE;
                         ids[n] = id; n++; return  SUCCESS; }
};

class SegList : public ArList<SegEntry> {
 public:
  GraySegsBuf  graySegsBuf;  // ids of entries be removed from R-tree
 public:
  void init () {}
  SegList () { init(); }
 public:
  int addSegment (Segment & seg);
 public:
  int createFromTextFile (const char * fileName);
  int print ();
};

class SegListPos : public ArListPos<SegEntry> {
 public:
  SegListPos ();
  SegListPos (SegList * segList) : ArListPos<SegEntry> (segList) {}
};

class SegListReader {
 private:
  SegList * segList;
  FILE *    fileP;
  char      fileName[256];
  int       curLineIdx;
  char      curLine[256];
 private:
  int       numPointSegments;
 public:
  void init ();
  SegListReader () { init(); }
  SegListReader (SegList * segList) { init(); this->segList = segList; }
  ~SegListReader () { if ( isFileOpened() ) closeFile(); }
 private:
  char * strContentPastEnd (char *);
  sbool isAsEmptyString (char *);
 private:
  void checkSegmentSize (sbool * isAccepted, Segment &);
 private:
  sbool isFileOpened () { return fileP != NULL; }
  int openFile (const char * fileName);
  int nextLineStep_read ();
  int nextLineStep ();
  int readAll ();
  void closeFile ();
 public:
  int readFile (const char * fileName);
};

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/

#endif      /* __seglistH__ */






