// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

#ifndef  __arlistH__
#define  __arlistH__

#include <mgrtree/BasDef.h>

#undef EndOfIncludesSection

template <class T>
class List {
 public:
  int      n;   // num of members
  T *      first;
  T *      last;
 public:
  void init () { n = 0; first = last = NULL; }
  List () { init(); }
 public:
  sbool isEmpty () { return  n == 0; }
 public:
  int addMember ();
  int addMember (T & memb) { int status = addMember ();
    if ( SUCCESS != status ) return FAILURE;
    else { *last = memb; last->next = NULL; return SUCCESS; }
  }
  void removeAll ();
 private:
  void bubblesort_singleRun (sbool * swapMade);
 public:
  void bubblesort ();
 public:
  ~List () { removeAll(); }
};

class ListPosStatus {
 public:
  short   o;
 public:
  ListPosStatus () { setPreStart(); }
 public:
  static short preStart () { return -1; }
  static short pastEnd () { return 1; }
  static short inside () { return 0; }
  void setInside () { this->o = inside(); }
  void setPreStart () { this->o = preStart(); }
  void setPastEnd () { this->o = pastEnd(); }
  sbool isInside () { return this->o == inside(); }
  sbool isPreStart () { return this->o == preStart(); }
  sbool isPastEnd () { return this->o == pastEnd(); }
};


template <class T>
class ListPos {
 private:
  List<T> *        list;
 public:
  ListPosStatus    status;
  T *              memb;
 public:
  void init () { list = NULL; memb = NULL; }
  void init (List<T> * list) { init(); this->list = list; }
  ListPos () { init(); }
  ListPos (List<T> * list) { init (list); }
  ListPos (ListPos<T> & pos) { *this = pos; }
  ~ListPos () {}
 public:
  int n() { return list->n; }
  void setPreStart () { status.setPreStart(); memb = NULL; }
  int nextStep ();
 public:
  sbool isLast () { return status.isInside() && memb == list->last; }
  ListPos<T> next () { ListPos<T> pos = *this; pos.nextStep(); return pos; }
  int swap (ListPos<T> & pos);
};

template <class T> class ArListNode;

template <class T>
class ArListNodeHeader {
 public:
  int              idx;        // index in list
  short            nEntries;   // # of (valid) entries in node
  short            unused;
  ArListNode<T> *  next;
 public:
  void init () {
    idx = -1; nEntries = 0; unused = 0; next = NULL; }
  ArListNodeHeader () { init(); }
 public:
};

#define MG_ArList_MaxNodeSize     1024  // in bytes
#define MG_ArList_MaxNodeEntries(T) \
  ((MG_ArList_MaxNodeSize - sizeof (ArListNodeHeader<T>)) / sizeof (T))

// memory chunk of node data
template <class T>
class ArListNode : public ArListNodeHeader<T> {
 public:
  T   entries [MG_ArList_MaxNodeEntries (T)];
 public:
  T & entry (short i) { return entries[i]; }
  short maxNumEntries () { return MG_Ca_size (entries); }
  sbool isEmpty () { return nEntries == 0; }
  sbool isFull () { return nEntries == maxNumEntries(); }
  T * lastMembP () {return isEmpty() ? NULL : &entry (nEntries - 1); }
  int add (T &);
};

template <class T>
class ArListNodes : public List <ArListNode<T> > {
};

// arrayed list
template <class T>
class ArList {
 public:
  int              n;  // num of all (valid) entries
  ArListNodes<T>   nodes;
 public:
  void init () { n = 0; }
  ArList () { init(); }
 public:
  sbool isEmpty () { return n == 0; }
  T * lastMembP () { return isEmpty() ? NULL : nodes.last->lastMembP(); }
  int add (T &);
};

template <class T>
class ArListPos {
 private:
  ArList<T> *               list;
 public:
  ListPosStatus             status;
  ArListNode<T> *           node;
  int                       inNodeIdx;
  T *                       memb;
 public:
  void init () { list = NULL; node = NULL; inNodeIdx = -1; memb = NULL; }
  void init (ArList<T> * list) { init(); this->list = list; }
  ArListPos () { init(); }
  ArListPos (ArList<T> * list) { init (list); }
  void init (T * memb);
  ArListPos (T * memb) { init (memb); }
  ~ArListPos () {}
 public:
  int n() { return list->n; }
  void setPreStart () {
    status.setPreStart(); node = NULL; memb = NULL; }
  int nextStep ();
 public:
  short order (ArListPos<T> & pos) {
    assert (status.isInside());
    return
    node->idx < pos.node->idx ? -1 : node->idx > pos.node->idx ? 1 :
    inNodeIdx < pos.inNodeIdx ? -1 : inNodeIdx > pos.inNodeIdx ? 1 : 0; }
};

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
int
List<T>::addMember ()
{
  T * memb = new T;
  if ( memb == NULL )
    return  FAILURE;
  memb->next = NULL;
  if ( last == NULL )
    first = last = memb;
  else {
    last->next = memb;
    if ( first == last )
      *first = *last;
  }
  last = memb;
  n++;
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
void
List<T>::removeAll ()
{
  T * current = first;
  while ( current != NULL ) {
    T * savedNext = current->next;
    delete (current);
    current = savedNext;
  }
  first = last = NULL;
  n = 0;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
void
List<T>::bubblesort_singleRun (
sbool * swapMade)
{
  ListPos<T> pos (this);
  *swapMade = FALSE;
  while ( SUCCESS == pos.nextStep() && !pos.isLast() ) {
    ListPos<T> next = pos.next();
    if ( pos.memb->order (*next.memb) == 1 ) {
      pos.swap (next);
      *swapMade = TRUE;
    }
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
void
List<T>::bubblesort ()
{
  sbool isSingleRunReq = TRUE;
  while ( isSingleRunReq ) {
    sbool swapMade;
    bubblesort_singleRun (&swapMade);
    isSingleRunReq = swapMade;
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
int
ListPos<T>::nextStep ()
{
  if ( status.isPastEnd() )
    return  FAILURE;
  if ( status.isPreStart() )
    memb = list->first;
  else {
    assert (memb != NULL);
    memb = memb->next;
  }
  if ( memb == NULL )
    status.setPastEnd();
  else
    status.setInside();
  return  status.isInside() ? SUCCESS : FAILURE;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
int
ListPos<T>::swap (
ListPos<T> & pos)
{
  if ( !status.isInside() || !pos.status.isInside() )
    return  FAILURE;
  MG_swap (T, *memb, *pos.memb);
  MG_swap (T*, memb->next, pos.memb->next); // swap back the next pointers
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
int
ArListNode<T>::add (
T & entry)
{
  if ( isFull() )
    return  FAILURE;
  short  i = nEntries;
  this->entry(i) = entry;
  this->entry(i).inNodeIdx = i;
  nEntries++;
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
int
ArList<T>::add (
T & entry)
{
  if ( nodes.isEmpty() || nodes.last->isFull() ) {
    int status = nodes.addMember();
    if ( SUCCESS != status )
      return  FAILURE;
    nodes.last->idx = nodes.n - 1;
  }
  int  status =
  nodes.last->add (entry);
  if ( SUCCESS != status )
    return  status;
  n++;
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
void
ArListPos<T>::init (
T * memb)
{
  this->memb = memb;
  inNodeIdx = memb->inNodeIdx;
  node = (ArListNode<T> *)
         ((char *) (memb - inNodeIdx) - sizeof (ArListNodeHeader <T>));
  status.setInside();
  list = NULL; //???
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
template <class T>
int
ArListPos<T>::nextStep ()
{
  if ( status.isPastEnd() )
    return  FAILURE;
  if ( status.isPreStart() ) {
    inNodeIdx = -1;
    node = list->nodes.first;
    status.setInside();
  }
 label_next:
  inNodeIdx++;

  if ( node == NULL ) {
    status.setPastEnd();
    memb = NULL;
    return  FAILURE;
  }
  if ( inNodeIdx >= node->nEntries ) {
    node = node->next;
    inNodeIdx = -1;
    goto label_next;
  }
  memb = &node->entries[inNodeIdx];
  return  SUCCESS;
}


#endif      /* __arlistH__ */






