// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

#ifndef  __MGRTREE_rtreeH__
#define  __MGRTREE_rtreeH__

#include <mgrtree/BasDef.h>
#include <mgrtree/CsizeStr.h>
#include <mgrtree/Assert.h>

#undef EndOfIncludesSection

#if !defined (MG_EI_Rt)
#define MG_EI_Rt MG_DllImport
#endif

class MG_EI_Rt XyPoint {
 public:
  double  x;
  double  y;
 public:
  XyPoint () {};
  XyPoint (double x, double y) : x(x), y(y) {}
};

#ifndef LONG_MIN
#define LONG_MIN  (-2147483647L - 1) /* minimum (signed) long value */
#define LONG_MAX	  2147483647L	/* maximum (signed) long value */
#endif

class XyRangeI;
class XyRangeD;

class MG_EI_Rt XyRange {
 public:
  XyPoint  ll;   // lower left  corner
  XyPoint  ur;   // upper right corner
 public:
  XyRange () {}
  XyRange (XyPoint & ll, XyPoint & ur) { fromLlUr (ll, ur); }
  void fromLlUr (XyPoint & ll, XyPoint & ur) {
    this->ll = ll; this->ur = ur; }
  void fromCoords (int * coords) {
    ll.x = coords[0]; ll.y = coords[1]; ur.x = coords[2]; ur.y = coords[3]; }
  XyRange (int * coords) { fromCoords (coords); }
  void toCoords (int * coords) {
    coords[0] = (int) ll.x; coords[1] = (int) ll.y;
    coords[2] = (int) ur.x; coords[3] = (int) ur.y; }
  void fromRngI (XyRangeI & r);
  XyRange (XyRangeI & r) { fromRngI (r); }
  XyRangeD & rngD () { return *((XyRangeD *)this); }
  XyRangeI rngI ();
 public:
  void setEmpty () { ll.x = ll.y = 0; ur.x = -1; ur.y = 1; }
  void setUnlimited () { ll.x = ll.y = LONG_MIN; ur.x = ur.y = LONG_MAX; }
 public:
  sbool isEmpty () const { return  ll.x > ur.x; }
  sbool isUnlimited () const { return  ll.x == LONG_MIN && ll.y == LONG_MIN
                                 && ur.x == LONG_MAX && ur.y == LONG_MAX; }
 public:
  const char * str ();
};

class UfXyRtreeRep;

//namespace XyRt {

// Xy R-tree record
class Rec {
};

#define MG_RtreeStrSize_RecId  16

// Record's Id. Either pointer or index.
union MG_EI_Rt RecId {
  Rec *  rec;
  long   idx;  // assert (sizeof(long) == sizeof(Rec *))
  RecId () {};
  RecId (void * rec) { this->rec = (Rec *) rec; };
  RecId (Rec * rec) { this->rec = rec; };
  RecId (int idx) { this->idx = idx; };
  void fromIdx (int idx) { this->idx = idx; }
  void setNull () { idx = -1; }
  sbool isNull () { return idx == -1; }
  sbool isEqual (RecId & recId) { return idx == recId.idx; }
  CsizeString<MG_RtreeStrSize_RecId> alpha ();
};

// Record Handle
class Rh {
 public:
  RecId  recId;
 public:
  Rh () {};
  void init (RecId & recId) { this->recId = recId; }
  Rh (RecId & recId) { init (recId); }
 public:
  virtual ~Rh () {}
  virtual XyRange range () = 0;
  virtual Rec * rec () { return recId.rec; } // default is pointer
};

class Searcher;

class SearcherRep {
 friend class Searcher;
 public:
  Searcher * owner;
  XyRange    searchRng;
 public:
  void init () { owner = NULL; searchRng.setEmpty(); }
  SearcherRep () { init(); }
 public:
  virtual ~SearcherRep () {};
  virtual void setRange (XyRange & range) = 0;

  // preferred way of the search iteration: while SUCCESS == nextStep
  virtual void setPreStart () = 0;
  virtual int nextStep () = 0;

  // more restrictive way (in user's point of view) to perform search:
  // in each step the state (including position) is saved using the call
  // stack
  virtual int processVisits ();
  virtual int visit () { return  FAILURE; }  // callback in processVisits,
                         // returns SUCCESS to continue with the process
};

class MG_EI_Rt CurRec {
 public:
  int              idx;     // count in iteration
  RecId            recId;   // current record Id
  XyRange          recRng;  // rec range, probably as stored in rtree
 public:
  void init () { idx = -1; recId.setNull(); recRng.setEmpty(); }
};

class XyRtree;

class MG_EI_Rt Searcher {
 public:
  SearcherRep * rep;
  CurRec        cur; // current info in searcher
 public:
  void init () { rep = NULL; cur.init(); }
  Searcher () { init(); }
  void setRep (SearcherRep * rep);
  Searcher (SearcherRep * rep) { init(); setRep (rep); }
  void init (XyRtree &);
  Searcher (XyRtree & rtree) { init (rtree); }
 public:
  ~Searcher () { if ( rep != NULL ) delete (rep); }
 public:
  void setRange (XyRange & r) { rep->setRange (r); }
  void setPreStart () { rep->setPreStart(); }
  int nextStep () { return rep->nextStep(); }
  int processVisits () { return rep->processVisits(); }
  int visit () { return rep->visit(); }
 public:
  int processVisits (XyRange & r) { setRange (r); return processVisits(); }
 public:
  XyRange & searchRng () { return rep->searchRng; }
 public:
  void expropriate () { init(); }
  Searcher & operator= (Searcher &);
};

class XyRtreeRep {
 public:
  XyRtree *         owner;
 public:
  Rh * rh (RecId & recId) { rh()->recId = recId; return rh(); }
 protected:
  void init () { owner = NULL; }
  XyRtreeRep () { init(); }
 public:
  virtual ~XyRtreeRep () {};
  virtual Rh * rh (); // default is owner's rhPrototype
  virtual void removeAll () = 0;
  virtual int insertRecord (RecId &) = 0;
  virtual int deleteRecord (RecId & recId) { return FAILURE; }
  virtual SearcherRep * newSearcherRep () = 0;
 public:
  virtual void printLevel (short levelIdx) {}
  virtual void printLevels (short topLevelIdx = 0,
                            short bottomLevelIdx = -1) {};
 public:
  int insertRecord (int idx) {
    RecId recId (idx); return insertRecord (recId); }
  int deleteRecord (int idx) {
    RecId recId (idx); return deleteRecord (recId); }
  int createSearcher (Searcher & searcher);
  int createSearcher (Searcher & searcher, XyRange & searchRng);
 public:
//  sbool hasOwner () { return owner != NULL; }
};

class MG_EI_Rt XyRtree {
                             friend class Searcher;
 protected:
  XyRtreeRep *   rep;
 public:
  Rh *           rhPrototype;   // possible to set record handle
                                // prototype outside the rep
  XyRange        overallRange;
  int            numRecords;
 public:
  Rh * rh () { return rep->rh(); }
  Rh * rh (RecId & recId) { return rep->rh(recId); }
 public:
  void init ();
  XyRtree () { init(); }
  void setRep (XyRtreeRep * rep);
  XyRtree (XyRtreeRep * rep) { init(); setRep (rep); }
 public:
  void setUfRep ();
 public:
  ~XyRtree ();
 public:
  void destroyRep ();
 public:
  void removeAll () { rep->removeAll(); }
 public:
  int insertRecord (RecId & recId) { return rep->insertRecord (recId); }
  int insertRecord (int idx) { RecId recId (idx);
                               return insertRecord (recId); }
  int deleteRecord (RecId & recId) { return rep->deleteRecord (recId); }
  int deleteRecord (int idx) { RecId recId (idx);
                               return deleteRecord (recId); }
 public:
  int createSearcher (Searcher & searcher) {
                             return rep->createSearcher (searcher); }
  int createSearcher (Searcher & searcher, XyRange & rng) {
                             return rep->createSearcher (searcher, rng); }
 public:
  void printLevel (short levelIdx) { rep->printLevel (levelIdx); }
  void printLevels (short topLevelIdx = 0, short  bottomLevelIdx = -1) {
    rep->printLevels (topLevelIdx, bottomLevelIdx); }
 public:
  void printSearch (Searcher & searcher);
  void printSearch (Searcher & searcher, XyRange & rng) {
    searcher.setRange (rng); printSearch (searcher); }
  void printSearch (Searcher & searcher, XyPoint & ll, XyPoint & ur) {
    printSearch (searcher, XyRange (ll, ur)); }
  void printSearch (XyRange & rng) { printSearch (Searcher(*this), rng); }
  void printSearch (XyPoint & ll, XyPoint & ur) {
    printSearch (Searcher(*this), ll, ur); }
 public:
  void expropriate () { init (); }
  XyRtree & operator= (XyRtree &);
};

//}  // end of namespace XyRt
//#ifndef MG_No_XyRt_namespace
//using namespace XyRt;
//#endif

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/

#endif      /* __MGRTREE_rtreeH__ */






