// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "xxdef.h"

#include <mgrtree/BasDef.h>
#include <mgrtree/System.h>
#include <mgrtree/Rtree.h>

#include <mgrtree/Assert.h>

#undef EndOfIncludesSection

/*------------------------------------------------------------------------- o
       private definitions
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
       static global variables
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
XyRtree::init ()
{
  rep = NULL;
  rhPrototype = NULL;
  overallRange.setEmpty();
  numRecords = 0;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
XyRtree::setRep (
XyRtreeRep * rep)
{
  assert (rep != NULL);
  assert (rep->owner == NULL);
  this->rep = rep;
  rep->owner = this;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
XyRtree::~XyRtree ()
{
  if ( rep == NULL ) {
    assert (rhPrototype == NULL);
    return;
  }
  if ( rhPrototype != NULL )
    delete rhPrototype;
  delete rep;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
XyRtree::destroyRep ()
{
  if ( rep != NULL )
    delete rep;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
XyRtree &
XyRtree::operator= (
XyRtree & rtree)
{
  assert (rep == NULL && rhPrototype == NULL);
  memcpy (this, &rtree, sizeof (*this));
  rtree.expropriate();
  return *this;
}

/*------------------------------------------------------------------------- o
  Alphabetic mapping of recId.
0-->"a", 1-->"b", ..., 25-->"z",26->"aa", ..., 26*26-1->"zz", 26*26->"aaa",
(null-->" ")
o -------------------------------------------------------------------------*/
CsizeString<MG_RtreeStrSize_RecId>
RecId::alpha ()
{
  CsizeString<MG_RtreeStrSize_RecId>  s;
  int alphabetN = 26; // abcdefghijklmnopqrstuvwxyz
  int i = idx;
  int d;

  if ( isNull() ) {
    s += ' ';
    return s;
  }
  if ( idx == 0 )
    s += 'a';

  while ( i > 0 ) {
    d = i % alphabetN;
    i /= alphabetN;
    s += 'a' + d;
  }
  s.swap();
  return  s;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
SearcherRep::processVisits ()
{
  setPreStart ();
  while ( SUCCESS == nextStep() ) {
    int status = visit();
    if ( SUCCESS != status )
      return  status;
  }
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Searcher::setRep (
SearcherRep * rep)
{
  assert (this->rep == NULL);
  assert (rep->owner == NULL);
  this->rep = rep;
  rep->owner = this;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
Searcher &
Searcher::operator= (
Searcher & searcher)
{
  assert (rep == NULL);
  memcpy (this, &searcher, sizeof (*this));
  searcher.expropriate();
  return *this;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Searcher::init (
XyRtree & rtree)
{
  init ();
  assert (rtree.rep != NULL);
  rtree.createSearcher (*this);
}


/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
Rh *
XyRtreeRep::rh ()
{
  return owner->rhPrototype;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
XyRtreeRep::createSearcher (
Searcher &  searcher)
{
  SearcherRep * searcherRep = newSearcherRep ();
  if ( searcherRep == NULL )
    return  FAILURE;
  searcher.setRep (searcherRep);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
XyRtreeRep::createSearcher (
Searcher &  searcher,
XyRange &   searchRng)
{
  int status = createSearcher (searcher);
  if ( SUCCESS != status )
    return  status;
  searcher.setRange (searchRng);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
XyRtree::printSearch (
Searcher & searcher)
{
  searcher.setPreStart();

  printf ("\nSearch result for range %s :", searcher.searchRng().str());
  while ( SUCCESS == searcher.nextStep () ) {
    printf ("\n%s: %s", searcher.cur.recId.alpha().s(),
                            searcher.cur.recRng.str());
  }
}



