// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
#include <string.h>
#include <stdio.h>
#include <malloc.h>

#include "xxdef.h"

#include <mgrtree/BasDef.h>
#include <mgrtree/System.h>
#include "UfRep.h"

#include <mgrtree/Assert.h>

#undef EndOfIncludesSection
/*------------------------------------------------------------------------- o
       private definitions
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
       static global variables
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
RtreeStatistics::init ()
{
  nCreatedNodes = 0;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
UfXyRtreeRep::init ()
{
  nLevels = 0;
//  maxNodeMembs = 4; // test only
  maxNodeMembs = 10;
  minNodeMembs = maxNodeMembs / 2;
  numRecords = 0;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
SearcherRep *
UfXyRtreeRep::newSearcherRep ()
{
  return  new UfSearcherRep (this);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::init ()
{
  rng.setEmpty();
  id.setNull();
  rtree = NULL;
  root = NULL;
  node = NULL;
  idx = -1;
  n = 0;
  levelIdx = -1;
  parent.rng.setEmpty();
  parent.node = NULL;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::init (
UfXyRtreeRep * rtree,
Root *         root)
{
  if ( root == NULL )
    root = &rtree->root;
  rng = root->rng;
  id.node = root->node;
  this->rtree = rtree;
  this->root = root;
  node = NULL;
  idx = 0;
  n = 1;
  if ( root->isEmpty() )
    n = 0;
  levelIdx = root->levelIdx;
  parent.rng.setEmpty();
  parent.node = NULL;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::rootStep ()
{
  init (rtree, root);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::rtreeRootStep ()
{
  this->root = &rtree->root;
  rootStep ();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::sepRootStep (
Root * sepRoot)
{
  RecId savedRecId = id.recId();
  root = sepRoot;
  sepRoot->rng = rng;
  sepRoot->node = isLeaf() ? NULL : id.node;
  sepRoot->levelIdx = levelIdx;
  rootStep ();
  if ( isLeaf() )
    id.recId() = savedRecId;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::setNewLeaf (
RecId & recId)
{
  assert (hasSeparateRoot());
  rng = rtree->rh(recId)->range().rngI();
  this->recId() = recId;
  node = NULL;
  idx = 0;
  n = 1;
  levelIdx = rtree->nLevels - 1;
  parent.node = NULL;
  parent.rng.setEmpty();
  root->rng = rng;
  root->node = NULL;
  root->levelIdx = levelIdx;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::upStep ()
{
  if ( isRoot() )
    return  FAILURE;
  sbool wasRootChild = isRootChild();
  id.node = node;
  levelIdx--;
  if ( wasRootChild ) {
    node = NULL;
    idx = 0;
    n = 1;
    rng = root->rng;
    parent.rng.setEmpty();
    return  SUCCESS;
  }
  idx = node->inParentIdx;
  node = node->parent;
  n = node->nMembers;
  if ( isRootChild() ) {
    parent.rng = root->rng;
    parent.node = NULL;
  }
  else {
    parent.node = node->parent;
    parent.getRange (node->inParentIdx);
  }
  rng = node->rng (idx);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::downStep (
short  o)
{
  if ( isLeaf() )
    return  FAILURE;
  if ( isRoot() ) {
    if ( root->node == NULL )
      return  FAILURE;
    parent.node = NULL;
    parent.rng = root->rng;
    node = root->node;
  }
  else {
    parent.node = node;
    parent.rng = rng;
    node = id.node;
  }
  n = node->nMembers;
  idx = -1; // out; means rng not valid yet
  levelIdx++;
  endStep (o);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::hidxStep (
short  hidx)
{
  int    contextN = horizForm.hasMaxContext ? rtree->maxNodeMembs : n;
  if ( !(0 <= hidx && hidx < contextN) ) {
    if ( horizForm.isOutable ) {
      idx = hidx;
      setEmpty();
    }
    return  FAILURE;
  }
  if ( idx == hidx )
    return  SUCCESS;  // already here
  idx = hidx;
  if ( hidx >= n ) {  // out of valid members
    setEmpty ();
    return  SUCCESS;
  }
  entryValue() = entry();

  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::pastEndStep ()
{
  short   o = 1;
  assert (horizForm.isOutable);
  assert (n < rtree->maxNodeMembs);
  pos().setEnd(o);
  hidxStep (idx + o);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
  skelet ... driven by nest acceptation only
o -------------------------------------------------------------------------*/
int
Member::nestStep_skelet (
NestStepDriver & driver)
{
  short & o = driver.o;
  short & ho = driver.ho;

  if ( driver.isPreStart ) {
    if ( rtree->isEmpty() )
      return  FAILURE;
    ho = -o;
    driver.isPreStart = FALSE;
    driver.getStatus (*this);
    return  SUCCESS;
  }

  if ( o == ho || ho == 0 ) {
    if ( SUCCESS == fastNeiStep (o) ) {
      ho = -ho;          // pre-children
      driver.getStatus (*this);
      return  SUCCESS;
    }
    else {
      ho = o;
      int   status = upStep();
      driver.getStatus (*this);
      return  status; // might return FAILURE
    }
  }
  else {
    if ( !driver.isNestAccepted ) {
      driver.blockDownStep();  // ho = -ho
      driver.getStatus (*this);
      return  SUCCESS;
    }
    downStep (-o);
    ho = isLeaf() ? 0 : -o;     // pre-children
    driver.getStatus (*this);
  }
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::nestStep (
NestStepDriver & driver)
{
 label_next:
  if ( SUCCESS != nestStep_skelet (driver) )
    return  FAILURE;
  if ( !driver.isAccepted )
    goto label_next;
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::rangeSearchLeafStep (
XyRangeI & searchRng)
{
  RangeSearchStepDriver driver (searchRng);
  if ( !isRoot() ) {
    assert (isLeaf());
    driver.isPreStart = FALSE;
    driver.ho = 0;
  }
  return
  nestStep (driver);  // o == 1
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::swap (
Member & memb)
{
  assert (levelIdx == memb.levelIdx);
  assert (!isRoot() && !memb.isRoot());

  MG_swap (NodeEntry, entryValue(), memb.entryValue());
  MG_swap (NodeEntry, entry(), memb.entry());

  if ( !isLeaf() ) {
    setChildrenParent();
    memb.setChildrenParent();
  }

}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::takeover (
Member & memb)
{
  assert (isEmpty());
  assert (!isRoot());
  assert (levelIdx == memb.levelIdx);

  entryValue() = memb.entryValue();
  entry() = entryValue();

  memb.setEmpty();
  if ( memb.isRoot() ) {
    assert (!isSibling (memb));
    assert (idx == n);
    memb.root->setEmpty();
    n++;
    node->nMembers = n;
  }
  else {
    assert (FALSE);
  }

  setChildrenParent();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::setRng ()
{
  if ( isRoot() )
    root->rng = rng;
  else
    entry().rng = entryValue().rng;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::setParentRng ()
{
  assert (!isRoot());
  parentMemb().setRng (parent.rng);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::adjustParentRng ()
{
  assert (!isRoot());
  Member  parent = parentMemb();
  parent.rng = rangesUnion();
  parent.setRng ();
  this->parent.rng = parent.rng;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
XyRangeI
Member::rangesUnion ()
{
  Member    memb = start();
  XyRangeI  unionRange = memb.rng;
  memb.horizForm.hasMaxContext = FALSE;

  while ( SUCCESS == memb.fastNextStep() ) {
    unionRange.applyUnion (memb.rng);
  }

  return  unionRange;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
Node *
Member::createEmptyNode ()
{
  short   entrySize = sizeof (NodeEntry);
  short   nodeSize = sizeof (NodeHeader) +
                     rtree->maxNodeMembs * entrySize;
  Node *  emptyNode = (Node *) malloc (nodeSize);
  emptyNode->init();
  int    i, n = rtree->maxNodeMembs;
  for (i = 0; i < n; i++) {
    emptyNode->entry(i).setEmpty();
  }
  return  emptyNode;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::createParent ()
{
  assert (isRoot());

  Node * newNode = createEmptyNode ();

  newNode->nMembers = 0;
  newNode->statisticId() = rtree->statistics.nCreatedNodes++;

  if ( !isLeaf() && rtree->nLevels != 0 ) {
    assert (!id.isNull());
    newNode->nMembers = 1;
    newNode->entry(0).id.node = id.node;
    id.node->parent = newNode;
    id.node->inParentIdx = 0;
  }
  node = newNode;
  entry() = entryValue();
  if ( levelIdx == 0 ) { // rtree's height to be increased
    if ( rtree->nLevels == 0 )
      rtree->nLevels = 2;
    else
      rtree->nLevels++;
    levelIdx = 1;
  }
  root->rng = rng;
  root->node = newNode;
  root->levelIdx = levelIdx - 1;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::findRecord (
RecId & recId)
{
  RangeSearchStepDriver driver (rtree->rh(recId)->range().rngI());
  rootStep ();
  while ( SUCCESS == nestStep (driver) ) {
    if ( this->recId().isEqual (recId) )
      return  SUCCESS;
  }
  return  FAILURE;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::checkChildParentReference ()
{
  if ( isRoot() || isLeaf() )
    return  SUCCESS;
  if ( id.node->parent != node )
    return  FAILURE;
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Member::checkRange ()
{
  if ( isLeaf() ) //???
    return  SUCCESS;
  Member child = startChild();
  return  rng.isEqual (child.rangesUnion()) ? SUCCESS : FAILURE;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
CsizeString<MG_RtreeStrSize_Node>
Member::cstr_node (
sbool  areIdsReq)
{
  CsizeString<MG_RtreeStrSize_Node>  s;
  char                               s0[256];
  if ( node == NULL ) {
    s = "<null node>";
    return  s;
  }
/*
  sprintf (s0, "%c:%d|%d: ",'a' + node->statisticId(),
                            (int) node->inParentIdx, (int) node->nMembers);
*/
//  sprintf (s0, "%c: ",'a' + node->statisticId());
  sprintf (s0, "%d: ", node->statisticId());

  s += s0;
  int   i, n = node->nMembers;
  for (i = 0; i < n; i++) {
    char *  sep = i == 0 ? "" : ",";
    s += sep;
    if ( areIdsReq ) {
      char  idStr[16];
      strcpy (idStr, "");
      if ( isLeaf() )
        sprintf (idStr, "%s", node->entry(i).recId().alpha().s());
      else {
        Node * childrenNode = node->entry(i).id.node;
        sprintf (idStr, "%d",
                 childrenNode == NULL ? -1 : childrenNode->statisticId());
      }
      s += idStr;
    }
    else {
      XyRangeI      rng = node->entry(i).rng;
      s += rng.cstr();
    }
  }
  return  s;
}

/*------------------------------------------------------------------------- o
   easy VC debug access
o -------------------------------------------------------------------------*/
const char *
Member::str_node ()
{
  static CsizeString<MG_RtreeStrSize_Node> s;
  s = cstr_node();
  return s;
}

/*------------------------------------------------------------------------- o
   easy VC debug access
o -------------------------------------------------------------------------*/
const char *
Member::str_node2 ()
{
  static CsizeString<MG_RtreeStrSize_Node> s;
  s = cstr_node (TRUE);
  return s;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
class CheckFlag {
 public:
  char isChecked;
 public:
  CheckFlag () { isChecked = FALSE; }
};

class FreedNodes {
 public:
  int         n;
  CheckFlag * flags;
 public:
  void init () { n = 0; flags = NULL; }
  FreedNodes () { init(); }
  void create (int n) { this->n = n; flags = new CheckFlag[n]; };
  ~FreedNodes () { if ( flags != NULL ) delete[] flags; }
 public:
  void check (int i) { if ( flags == NULL ) return;
                       flags[i].isChecked = TRUE; }
  void printStatus () {
    for (int i = 0; i < n; i++)
      if ( !flags[i].isChecked )
        printf ("\nFreedNodesChecker: %d not freed", i);
  }
};

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Member::destroyDescendantNodes ()
{
  if ( root->isEmpty() )
    return;
  assert (isRoot());
  DestroyChildrenNodeStepDriver driver;

//  sbool                         isCheckReq = TRUE;
  sbool                         isCheckReq = FALSE;
  FreedNodes                    freedNodes;
  if ( hasSeparateRoot() )
    isCheckReq = FALSE;
  if ( isCheckReq )
    freedNodes.create (rtree->statistics.nCreatedNodes);

  while ( SUCCESS == nestStep (driver) ) {
    if ( isCheckReq )
      freedNodes.check (id.node->statisticId());
    assert (id.node != NULL);
    free (id.node);
    id.node = NULL;
  }
  root->setEmpty();
  if ( isCheckReq )
    freedNodes.printStatus ();
  if ( !hasSeparateRoot() )
    rtree->setEmpty();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
Member
UfXyRtreeRep::firstRootChild ()
{
  Member  memb (this);
  memb.downStep();
  return  memb;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
UfXyRtreeRep::setEmpty ()
{
  nLevels = 0;
  root.setEmpty();
  numRecords = 0;
  statistics.init();
  if ( owner != NULL ) {
    owner->overallRange.setEmpty();
    owner->numRecords = 0;
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
UfXyRtreeRep::destroyAllNodes ()
{
  Member (this).destroyDescendantNodes();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
UfXyRtreeRep::printLevel (
short  levelIdx)
{
  LevelStepDriver  driver (levelIdx);
  Member           memb (this);
  int              indentSize = 4;
  int              curLineLen = indentSize;
  int              maxLineLen = 80;
  if ( levelIdx == 0 )
    printf ("\n");
  printf ("L%d: ", levelIdx);
  while ( SUCCESS == memb.nestStep (driver) ) {
    if ( !memb.isStart() )
      continue;
    const char * s = memb.cstr_node (TRUE).s();
    curLineLen += strlen (s) + 2;
    if ( curLineLen > maxLineLen ) {
      curLineLen = indentSize + strlen (s) + 2;
      printf ("\n    ");
    }
    printf ("  %s", s);
  }
  if ( levelIdx == 0 ) {
    printf ("  #records: %d (a-%s), #created nodes: %d",
            numRecords,
            RecId (numRecords-1).alpha().s(),
            statistics.nCreatedNodes);
  }
  printf ("\n");
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
UfXyRtreeRep::printLevels (
short  topLevelIdx,
short  bottomLevelIdx)
{
  if ( bottomLevelIdx == -1 )
    bottomLevelIdx = nLevels - 1;
  int   i, n = bottomLevelIdx - topLevelIdx + 1;
  printf ("\n");
  for (i = 0; i < n; i++)
    printLevel (topLevelIdx + i);
  printf ("\n");
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
UfSearcherRep::setPreStart ()
{
  memb.rootStep ();
  owner->cur.idx = -1;
  owner->cur.recId.setNull();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
UfSearcherRep::nextStep ()
{
  if ( SUCCESS != memb.rangeSearchLeafStep (searchRngI) )
    return  FAILURE;
  owner->cur.idx++;
  owner->cur.recId = memb.recId();
  owner->cur.recRng.fromRngI (memb.rng);
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
XyRtree::setUfRep ()
{
  UfXyRtreeRep * rep = new UfXyRtreeRep;
  setRep (rep);
}


