// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

#ifndef  __MGRTREE_ufrepH__
#define  __MGRTREE_ufrepH__

#include <mgrtree/BasDef.h>
#include <mgrtree/Rtree.h>
#include "XyRangeT.h"

#undef EndOfIncludesSection

class Node;

// assert (sizeof(NodeOrRecId) == 4)
union NodeOrRecId {
  Node *       node;
  int          idx;
  Rec * rec;
  NodeOrRecId () {};
  NodeOrRecId (Node * node) { this->node = node; };
  NodeOrRecId (int idx) { this->idx = idx; };
  void fromIdx (int idx) { this->idx = idx; }
  sbool isNull () { return idx == -1; }
  void setNull () { idx = -1; }
  RecId & recId () { return *(RecId *) this; }
};

// 8-byte header
class NodeHeader {
 public:
  short    reserved;
  char     nMembers;
  char     inParentIdx;
  Node *   parent;
 public:
  void init () {
    reserved = 0; nMembers = 0; inParentIdx = -1; parent = NULL; }
 public:
  short & statisticId () { return  reserved; }
};

class NodeEntry {
 public:
  XyRangeI     rng;
  NodeOrRecId  id;
 public:
  RecId & recId () { return id.recId(); }
  void setEmpty () { rng.setEmpty(); id.setNull(); }
};


// memory chunk of node data
class Node : public NodeHeader {
 public:
  NodeEntry   entries[4];  // 4 or more
 public:
  NodeEntry & entry (short i) { return entries[i]; }
  XyRangeI & rng (short i) { return entries[i].rng; }
};

class PosWn {
 public:
  int    idx;
  int    n;
 public:
  void init () { idx = -1; n = 0; }
  sbool isEnd (short o = 1) { return o == -1 ? idx == 0 : idx == n - 1; }
  sbool isOut (short o) { return o == -1 ? idx == -1 : idx == n; }
  void setEnd (short o = 1) { idx = o == -1 ? 0 : n - 1; }
  int neiStep (short o = 1) {
    int  neiIdx = idx + o;
    if ( !(0 <= neiIdx && neiIdx < n) ) return  FAILURE;
    idx = neiIdx;
    return  SUCCESS;
  }
  PosWn end (short o = 1) { PosWn  pos = *this; pos.setEnd(o); return pos; }
};

class Root {
 public:
  XyRangeI  rng;
  Node *    node;
  short     levelIdx; // 0 for rtree's root
 public:
  void init () { rng.setEmpty(); node = NULL; levelIdx = 0; }
  Root () { init(); }
 public:
  sbool isEmpty () { return node == NULL; }
  void setEmpty () { init(); }
};

class RtreeStatistics {
 public:
  int    nCreatedNodes;
 public:
  void init ();
  RtreeStatistics () { init(); }
 public:

};

class Member;

// UniForm R-tree Representation:
// (rng,id)-form for both leaf and non-leaf node entry.
// Ranges stored with integer coordinates.
class UfXyRtreeRep : public XyRtreeRep {
 public:
  short            nLevels;
  Root             root;
  short            maxNodeMembs;  // "node capacity"
  short            minNodeMembs;  // 2 <= minNodeMembs <= maxNodeMembs/2
  int              numRecords;
  RtreeStatistics  statistics;
 public:
  void init ();
  UfXyRtreeRep () { init (); }
  void setMaxNodeMembs (short value) {
    assert (numRecords == 0);
    maxNodeMembs = value; minNodeMembs = maxNodeMembs / 2; }
 public:
  void setEmpty ();
  sbool isEmpty () { return root.isEmpty(); }
 public:
  void destroyAllNodes ();
 public:
  virtual ~UfXyRtreeRep () { removeAll(); }
  virtual void removeAll () { destroyAllNodes (); }
  virtual int insertRecord (RecId &);
  virtual int deleteRecord (RecId &);
  virtual SearcherRep * newSearcherRep ();
 public:
  virtual void printLevel (short levelIdx);
  virtual void printLevels (short topLevelIdx = 0,
                            short bottomlevelIdx = -1);
 public:
  Member firstRootChild ();
  sbool hasSingleRecord (RecId &);
  void shorten ();
};

struct MemberParent {
  Node *         node;
  XyRangeI       rng;
  void getRange (short  idx) { rng = node->entries[idx].rng; }
};

class HorizForm {
 public:
  sbool hasMaxContext;
  sbool isOutable;
 public:
  void init () { hasMaxContext = FALSE; isOutable = FALSE; }
  HorizForm () { init(); }
};

class NestStepDriver;

#define MG_RtreeStrSize_Node  128

class Member : public NodeEntry {
 public:
  UfXyRtreeRep *  rtree;
  Root *          root;  // might differ from &rtree->root (separate root)
  Node *          node;
  int             idx;   // horizontal idx within members (siblings)
  int             n;     // number of members
  int             levelIdx;
  HorizForm       horizForm;
  MemberParent    parent;
 public:
  NodeEntry & entryValue () { return *this; }
  NodeEntry & entry () { return node->entries[idx]; }
 public:
  void init ();
  void init (UfXyRtreeRep * rtree, Root * root = NULL);
  Member () { init(); }
  Member (Member * memb) { *this = *memb; }
  Member (Member & memb) { *this = memb; }
  Member (UfXyRtreeRep * rtree, Root * root = NULL) { init (rtree, root); }
 public:
//  sbool isValid () { return  rtree != NULL && node != NULL; }
//  void setInvalid () { node = NULL; levelIdx = -1; }
  short leafLevelIdx () { return rtree->nLevels - 1; }
  sbool isEmpty () { return rng.isEmpty(); }
  sbool isRoot ()       { return node == NULL; }
  sbool isRootChild ()  { return root->node == node && !isRoot(); }
  sbool isLeaf ()       { return levelIdx == leafLevelIdx(); }
 public:
  PosWn & pos() { return  (PosWn &) idx; }
  sbool isEnd (short o = 1) { return pos().isEnd(o); }
  sbool isStart () { return isEnd(-1); }
  sbool isSibling (Member & memb) {
    return !isRoot() && node == memb.node; }
  sbool isOut (short o) { return pos().isOut(o); }
 public:
  void setEmpty () { entryValue().setEmpty(); }
 public:
  void swap (Member & memb);
  void takeover (Member &);
  void patchEmptyWithEnd () { swap (end()); }
 public:
  void setRng ();
  void setRng (XyRangeI & r) { rng = r; setRng(); }
  void setParentRng ();
  void adjustParentRng ();
  XyRangeI rangesUnion ();
  void setChildrenParent () { if ( isLeaf() || id.isNull() ) return;
    id.node->parent = node;
    id.node->inParentIdx = idx;
  }
 public:
  int hidxStep (short hidx);
  int endStep (short o = 1) { return hidxStep (pos().end(o).idx); }
  int startStep () { return endStep (-1); }
  int pastEndStep ();
  int upStep ();
  int downStep (short o = -1);
  int neiStep (short o = 1) { return hidxStep (idx + o); }
  int nestStep_skelet (NestStepDriver &);
  int nestStep (NestStepDriver &);
 public:
  int fastNeiStep (short o) {
    if ( o == -1 && idx == 0 || o == 1 && idx == n - 1 )  return  FAILURE;
    idx += o;
    entryValue() = entry();
    return  SUCCESS;
  }
  int fastNextStep () {
    if ( idx == n - 1 )  return  FAILURE;
    else { idx++; entryValue() = entry(); return  SUCCESS; }
  }
  int fastPrevStep () {
    if ( idx == 0 )  return  FAILURE;
    else { idx--; entryValue() = entry(); return  SUCCESS; }
  }
 public:
  sbool hasSeparateRoot () { return  &rtree->root != root; }
 public:
  void rootStep ();
  void rtreeRootStep ();
  void sepRootStep (Root *);
  void setNewLeaf (RecId & recId);
 public:
  int rangeSearchLeafStep (XyRangeI & searchRng);
  int rangeSearchLeafStep (XyRange & searchRng) {
    return rangeSearchLeafStep (searchRng.rngI()); }
 public:
  Member nei (short o = 1) {
    Member memb(this); memb.neiStep(o); return memb; }
  Member end (short o = 1) {
    Member memb(this); memb.endStep(o); return memb; }
  Member start () { return  end(-1); }
  Member parentMemb () { Member memb(this); memb.upStep(); return memb; }
  Member startChild () { Member memb(this);
    if ( isLeaf() ) memb.setEmpty(); else memb.downStep(-1); return memb; }
 public:
  int findRecord (RecId &);
 public:
  Node * createEmptyNode ();
  void createParent ();
  void destroyDescendantNodes ();
 public:
  sbool hasFewSiblings ();
  sbool hasRangeBoundaryTouch (XyRangeI &);
  sbool hasRangeBoundaryTouch ();
  void remove ();

 public:
  CsizeString <MG_RtreeStrSize_Node> cstr_node (sbool  areIdsReq = FALSE);
  const char * str_node();
  const char * str_node2();
 public:
  int checkChildParentReference ();
  int checkRange ();
  void printChildrenRanges ();
};

class NestStepDriver {
 public:
  sbool   isPreStart;
  short   o;             // direction of nest iteration (usually 1)
  short   ho;            // horizontal orienation
                         // says if descendants have to be stepped
  sbool   isAccepted;
  sbool   isNestAccepted;
 public:
  void setAccepted () { isAccepted = isNestAccepted = TRUE; }
  void setRejected () { isAccepted = isNestAccepted = FALSE; }
  void blockDownStep () { ho = -ho; }
 public:
  NestStepDriver () { isPreStart = TRUE; o = 1; ho = -1; setAccepted(); }
 public:
  virtual void getStatus (Member &) { setAccepted(); }
};

class RangeSearchStepDriver : public NestStepDriver {
 public:
  XyRangeI searchRng;
  sbool    areLeafsOnlyReq;
 public:
  RangeSearchStepDriver (XyRangeI & searchRng) {
    this->searchRng = searchRng; areLeafsOnlyReq = TRUE; }
  RangeSearchStepDriver (XyRange & searchRng) {
    this->searchRng = searchRng.rngI(); areLeafsOnlyReq = TRUE; }
 public:
  virtual void getStatus (Member & memb) {
    if ( ho == o || !memb.rng.isOverlapping (searchRng) )  setRejected();
    else { isNestAccepted = TRUE;
           isAccepted = areLeafsOnlyReq ? memb.isLeaf() : TRUE; }
  }
};

class DestroyChildrenNodeStepDriver : public NestStepDriver {
 public:
  virtual void getStatus (Member & memb) {
    isAccepted = ho == 1;  // after children have beed processed
    isNestAccepted = memb.levelIdx < memb.leafLevelIdx() - 1;
  }
};

class InnerNodeStepDriver : public NestStepDriver {
 public:
  virtual void getStatus (Member & memb) {
    isAccepted = ho == -o &&  // before children have beed processed
                 !memb.isRoot();
    isNestAccepted = memb.levelIdx < memb.leafLevelIdx() - 1;
  }
};

class LevelStepDriver : public NestStepDriver {
 public:
  short    levelIdx;  // idx of level to traverse
 public:
  void init () { levelIdx = 0; }
  LevelStepDriver () { init(); }
  LevelStepDriver (short levelIdx) { init(); this->levelIdx = levelIdx; }
 public:
  virtual void getStatus (Member & memb) {
    isAccepted = ho != 1 && memb.levelIdx == levelIdx;
    isNestAccepted = memb.levelIdx < levelIdx;
  }
};

class FindByIdDriver : public NestStepDriver {
 public:
  short         isRecId;  // FALSE for chidrenId
  short         statisticId;
  RecId  recId;
 public:
  void init () { isRecId = FALSE; statisticId = -1; recId.setNull(); }
  FindByIdDriver () { init(); }
  FindByIdDriver (short statisticId) : statisticId (statisticId) {
    isRecId = FALSE; }
  FindByIdDriver (RecId & recId) : recId (recId) { isRecId = TRUE; }
 public:
  sbool isIdMemb (Member & memb) {
    if ( isRecId )  return  memb.recId().isEqual(recId);
    return  memb.isLeaf() ? FALSE :
            memb.id.node->statisticId() == statisticId;
  }
  virtual void getStatus (Member & memb) {
    isNestAccepted = TRUE;
    isAccepted = ho != 1 && isIdMemb (memb);
  }
};

class Inserter {
 public:
  UfXyRtreeRep * rtree;  // for convenience, equal to pri.rtree
  Member         pri;    // primary members, refers to node existing in
                         // tree
  Member         sec;    // secondary members, refers to orphan node to be
                         // inserted
  Root           secRoot;  // secondary root
 public:
  Member & prisec (short o) { return  o == -1 ? pri : sec; }
 public:
  Inserter (UfXyRtreeRep * rtree);
 private:
  short priSecO (Member & memb) { for (short o = -1; o <= 1; o += 2) {
    if ( memb.isSibling (prisec(o)) ) return o; } return 0; }
 private:
  int membPriSecNextStep (Member &);
  int getMaxAreaRemainderPredecessor (double &, Member &, Member &);
  int pickSeeds ();
  Member pickMember ();
  short pickedDistrib (Member & pickedMemb);
  short numRemaining () { return  pri.n - pri.idx; }
  short numToMinDemand ();
  int distribute (); // "quadratic split"
  void insertSibling ();
  void adjustAnchestorsRngs ();
  void chooseSmallestAreaIncrementMemb ();
  void chooseLeaf ();
 public:
  void insertRecord (RecId &);

};

class Deleter {
 public:
  UfXyRtreeRep * rtree;  //
  RecId          recId;
  Member         sep;    // member to separate
  Root           sepRoot;  // separate root
 public:
  void init () { rtree = NULL; }
  void init (UfXyRtreeRep * rtree) {
    this->rtree = rtree; sep.init (rtree); }
  Deleter (UfXyRtreeRep * rtree) { init (rtree); }
 private:
  int reinsertSepLeafs ();
 public:
  int deleteRecord (RecId &);
};

class UfSearcherRep : public SearcherRep {
 public:
  Member      memb;
  XyRangeI    searchRngI;  // rounded form
 public:
  void init () {
    memb.init (); searchRngI.setEmpty();}
  UfSearcherRep () { init(); }
  UfSearcherRep (UfXyRtreeRep * rtree) {
    init(); memb.init (rtree); }
 public:
  virtual ~UfSearcherRep () {};
  virtual void setRange (XyRange & rng) {
    searchRng = rng; searchRngI = rng.rngI(); }
  virtual void setPreStart ();
  virtual int nextStep ();
};

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/

#endif      /* __MGRTREE_ufrepH__ */






