// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
#include <string.h>
#include <stdio.h>
#include <malloc.h>

#include "xxdef.h"

#include <mgrtree/BasDef.h>
#include <mgrtree/System.h>
#include "UfRep.h"

#include <mgrtree/Assert.h>

#undef EndOfIncludesSection
/*------------------------------------------------------------------------- o
       private definitions
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
       static global variables
o -------------------------------------------------------------------------*/

/*------------------------------------------------------------------------- o
  nei step in pri+sec members
o -------------------------------------------------------------------------*/
int
Inserter::membPriSecNextStep (
Member &    memb)
{
  if ( SUCCESS == memb.fastNextStep() )
    return  SUCCESS;
  // isEnd()
  if ( memb.isSibling (sec) )
    return FAILURE;
  // go to the sec
  memb = sec;
  memb.startStep();
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Inserter::getMaxAreaRemainderPredecessor (
double & maxAreaRemainder,  // <= maximum value within predecessors
Member & maxMemb,           // <= choosen horizontal predecessor of
                            //    questMemb within pri+sec siblings
Member & questMemb)         // => member to get the result for
{
  Member  memb = questMemb.isSibling(sec) ? pri.end() : questMemb.nei(-1);
  double  questMembArea = questMemb.rng.area();
  sbool   isFirstTaken = TRUE;

  do {
    double  areaRemainder =
              questMemb.rng.unionAreaRemainder (questMembArea, memb.rng);
    if ( isFirstTaken || areaRemainder > maxAreaRemainder ) {
      maxAreaRemainder = areaRemainder;
      maxMemb = memb;
      isFirstTaken = FALSE;
    }
  }  while ( SUCCESS == memb.fastPrevStep () );

  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
Inserter::pickSeeds ()
{
  Member  seedMemb0, seedMemb1;
  Member  memb1;
  double  seedsAreaRemainder;
  sbool   isFirstTaken = TRUE;

  pri.startStep();
  sec.startStep();

  memb1 = pri;

  while ( SUCCESS == membPriSecNextStep (memb1) ) {
    double  areaRemainder;
    Member  memb0;
    getMaxAreaRemainderPredecessor (areaRemainder, memb0, memb1);
    if ( isFirstTaken || areaRemainder > seedsAreaRemainder ) {
      seedsAreaRemainder = areaRemainder;
      seedMemb0 = memb0;
      seedMemb1 = memb1;
      isFirstTaken = FALSE;
    }
  }
  // swap contents with the starts
  pri.swap (seedMemb0);
  sec.swap (seedMemb1);

  for (short o = -1; o <= 2; o += 2 ) {
    assert (prisec(o).isStart());
    prisec(o).parent.rng = prisec(o).rng; // set union range to seed's rng
    prisec(o).neiStep();
  }
  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
short
Inserter::numToMinDemand ()
{
  for (short o = -1; o <= 1; o += 2) {
    short numToMin = rtree->minNodeMembs - prisec(o).idx;
    assert (numRemaining() >= numToMin);
    if ( numRemaining() == numToMin )
      return o;
  }
  return 0;
}


/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
Member
Inserter::pickMember ()
{
  Member  pickedMemb = pri;
  Member  memb = pri;
  double  maxIncDifference = -1.0;
  double  priArea = pri.parent.rng.area();
  double  secArea = sec.parent.rng.area();

  while ( SUCCESS == memb.fastNextStep() ) {
    double  priInc, secInc, incDifference;
    priInc = pri.parent.rng.unionAreaIncrement (priArea, memb.rng);
    secInc = sec.parent.rng.unionAreaIncrement (secArea, memb.rng);
    incDifference = MG_abs (priInc - secInc);
    if ( incDifference > maxIncDifference ) {
      pickedMemb = memb;
      maxIncDifference = incDifference;
    }
  }
  return  pickedMemb;
}


/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
short           // <= pri/sec
Inserter::pickedDistrib (
Member  & pickedMemb)
{
  double  priInc = pri.parent.rng.unionAreaIncrement (pickedMemb.rng);
  double  secInc = sec.parent.rng.unionAreaIncrement (pickedMemb.rng);
  if ( priInc < secInc ) return  -1;
  if ( priInc > secInc ) return  1;

  double priArea = pri.parent.rng.area();
  double secArea = sec.parent.rng.area();
  if ( priArea < secArea ) return  -1;
  if ( priArea > secArea ) return  1;

  int  priNum = pri.idx;
  int  secNum = sec.idx;
  if ( priNum < secNum ) return  -1;
  if ( priNum > secNum ) return  1;

  return -1; // "either of {-1,1}"
}

/*------------------------------------------------------------------------- o
  quadratic split distribution
o -------------------------------------------------------------------------*/
int
Inserter::distribute ()
{
  pickSeeds ();
  assert (pri.idx == 1 && sec.idx == 1); // past start

  short          o;

//  if ( rtree->numRecords == 10 )
//    MgSystem::enterDebug ();
  while ( numRemaining() > 0 && numToMinDemand() == 0 ) {
    Member  pickedMemb = pickMember ();
    o = pickedDistrib (pickedMemb);
    Member & memb = prisec(o);
    memb.swap (pickedMemb);
    if ( o == 1 ) {              // maintain "nonholeyness"
      pickedMemb.patchEmptyWithEnd();
      if ( pickedMemb.idx == pri.idx )
        pri = pickedMemb;
      pri.n--;
    }
    memb.parent.rng.applyUnion (memb.rng);
    memb.neiStep();
  }

  if ( numRemaining() > 0 ) {
    o = numToMinDemand();
    assert (o != 0);
    // there are exactly numRemaining() members to satisfy the
    // minNodeMembs demand
    if ( o == -1 ) // remaining members on the place in primary part
      do {
        pri.parent.rng.applyUnion (pri.rng);
      }  while ( SUCCESS == pri.neiStep() );
    else {
      short savedIdx = pri.idx;
      do {
        sec.swap (pri);
        sec.parent.rng.applyUnion (sec.rng);
        sec.neiStep();
      }  while ( SUCCESS == pri.neiStep() );
      pri.idx = savedIdx;
    }
  }

  // adjust nMembers and parent ranges
  for (o = -1; o <= 1; o += 2) {
    short   n = prisec(o).idx;
    prisec(o).n = n;
    prisec(o).node->nMembers = (char) n;
//    prisec(o).adjustParentRng(); ... parent rng computed already
                                     // and is to be set only
    prisec(o).setParentRng();
  }

  return  SUCCESS;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Inserter::insertSibling ()
{
  if ( pri.isRoot() ) {
    pri.createParent ();
    sec.levelIdx = 1;
  }

  if ( pri.n == rtree->maxNodeMembs ) {
    sec.createParent();
    distribute ();
    pri.upStep();
    sec.upStep();
    return;
  }

  pri.pastEndStep();
  pri.takeover (sec);
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Inserter::adjustAnchestorsRngs ()
{
  sbool     isRngIncluded = FALSE;
  XyRangeI  rng;
  while ( !isRngIncluded && !pri.isRoot() ) {
    rng = pri.rangesUnion();
    pri.upStep();
    pri.rng = rng;
    pri.setRng ();
    isRngIncluded = pri.parent.rng.includes (pri.rng);
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Inserter::chooseSmallestAreaIncrementMemb ()
{
  Member memb = pri.start();
  double secArea = sec.rng.area();
  double smallestAreaInc = sec.rng.unionAreaIncrement (secArea, memb.rng);
  double smallestArea = memb.rng.area();         // for resolving ties

  while ( SUCCESS == memb.fastNextStep() ) {
    double areaInc = sec.rng.unionAreaIncrement (secArea, memb.rng);
    if ( areaInc > smallestAreaInc )
      continue;
    double area = memb.rng.area();
    if ( areaInc < smallestAreaInc || area < smallestArea ) {
      pri = memb;
      smallestAreaInc = areaInc;
      smallestArea = area;
    }
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Inserter::chooseLeaf ()
{
  if ( rtree->nLevels == 0 )
    return;
  while ( !pri.isLeaf() ) {
    if ( !pri.isRoot() )
      chooseSmallestAreaIncrementMemb ();
    pri.downStep();
  }
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
void
Inserter::insertRecord (
RecId & recId)
{
  sec.setNewLeaf (recId);

  chooseLeaf ();

  sbool  isLeafRngIncluded = pri.parent.rng.includes (sec.rng);

  while ( !sec.isEmpty() ) {
    insertSibling ();
  }
  if ( !isLeafRngIncluded )
    adjustAnchestorsRngs ();
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
Inserter::Inserter (
UfXyRtreeRep * rtree)
 : rtree (rtree)
{
  pri.init (rtree);
  pri.horizForm.isOutable = TRUE;
  sec.init (rtree, &secRoot);
  sec.horizForm.hasMaxContext = TRUE;
}

/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/
int
UfXyRtreeRep::insertRecord (
RecId & recId)
{
  Inserter  inserter (this);
  inserter.insertRecord (recId);
  numRecords++;
  if ( owner != NULL ) {
    owner->overallRange.fromRngI (root.rng);
    owner->numRecords = numRecords;
  }
  return  SUCCESS; //???
}



