// R-tree Library. A simple 2d- C++ implementation of the R-tree structure
//
// Copyright (C) 2002-2003  Ondrej Pavlata.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
// Contact information:
//  Mailing address:
//    Ondrej Pavlata
//    V. Nezvala 7a
//    466 02 Jablonec nad Nisou
//    Czech republic
//
//  Email:
//    o.j.p@worldonline.cz

#ifndef  __MGRTREE_xyrangetH__
#define  __MGRTREE_xyrangetH__

#include <math.h>

#include <mgrtree/BasDef.h>
#include <mgrtree/CsizeStr.h>
#include <mgrtree/Rtree.h>

#undef EndOfIncludesSection

template <class T>
class XyPointT {
 public:
  T       x;
  T       y;
 public:
  XyPointT () {};
  XyPointT (T x, T y) : x(x), y(y) {}
};

typedef XyPointT<int>    XyPointI;
typedef XyPointT<double> XyPointD;

#define MG_RtreeStrSize_XyRange  64

// rectangular (xy axis paralel) range
// interpreting  ll.x > ur.x as empty range
// assumed the coordinates domain is (LONG_MIN,LONG_MAX)^2
template <class T>
class XyRangeT {
 public:
  XyPointT<T>  ll;   // lower left  corner
  XyPointT<T>  ur;   // upper right corner
 public:
  XyRangeT () {};
  XyRangeT (XyPointT<T> & ll, XyPointT<T> & ur) { fromLlUr (ll, ur); }
  void fromLlUr (XyPointT<T> & ll, XyPointT<T> & ur) {
    this->ll = ll; this->ur = ur; }
 public:
  void setEmpty () { ll.x = 0; ll.y = 0;
                     ur.x = -1; ur.y = 1; } // area is -1.0
  void setUnlimited () { ll.x = ll.y = LONG_MIN;
                         ur.x = ur.y = LONG_MAX; }
 public:
  sbool isEmpty () const { return  ll.x > ur.x; }
  sbool isUnlimited () const { return  ll.x == LONG_MIN && ll.y == LONG_MIN
                                 && ur.x == LONG_MAX && ur.y == LONG_MAX; }
  sbool isEqual (const XyRangeT<T> & r) const {
    return ll.x == r.ll.x && ll.y == r.ll.y &&
           ur.x == r.ur.x && ur.y == r.ur.y; }
 public:
  void applyPoint_rect (const XyPointT<T> & p) {
    if ( p.x < ll.x ) ll.x = p.x; else if ( p.x > ur.x ) ur.x = p.x;
    if ( p.y < ll.y ) ll.y = p.y; else if ( p.y > ur.y ) ur.y = p.y;
  }
  void applyPoint (const XyPointT<T> & pnt) {
    if ( isEmpty() ) ll = ur = pnt;
    else applyPoint_rect (pnt);
  }
  XyRangeT<T> unionRect (const XyRangeT<T> & r) { XyRangeT<T> u = r;
    u.applyPoint_rect (ll); u.applyPoint_rect (ur); return u; }
  XyRangeT<T> unionXyRange (const XyRangeT<T> & r) {
    if ( isEmpty() ) return r;
    else if ( r.isEmpty() ) return *this;
    else return unionRect (r);
  }
  void applyUnion (const XyRangeT<T> & r) { *this = unionXyRange (r); }
 public:
  double area () const {
    return  ((double)ur.x - ll.x) * ((double)ur.y - ll.y); }
  double unionArea (const XyRangeT<T> & r) {
    return  unionRect(r).area(); }
  double unionAreaIncrement (const XyRangeT<T> & r) {
    return  unionArea(r) - area(); }
  double unionAreaIncrement (double myArea, const XyRangeT<T> & r) {
    return  unionArea(r) - myArea; }
  double unionAreaRemainder (const XyRangeT<T> & r) {
    return  unionArea(r) - area() - r.area(); }
  double unionAreaRemainder (double myArea, const XyRangeT<T> & r) {
    return  unionArea(r) - myArea - r.area(); }

 public:
  sbool isOverlapping (const XyRangeT<T> & r) const {
    if ( isEmpty() || r.isEmpty() ) return  FALSE;
    return  ll.x <= r.ur.x && r.ll.x <= ur.x &&
            ll.y <= r.ur.y && r.ll.y <= ur.y; }
  sbool includes (const XyRangeT<T> & r) const {
    if ( r.isEmpty() ) return TRUE; else if ( isEmpty() ) return  FALSE;
    return  ll.x <= r.ll.x && r.ur.x <= ur.x &&
            ll.y <= r.ll.y && r.ur.y <= ur.y; }
  sbool includes_inner (const XyRangeT<T> & r) const {
    return  ll.x < r.ll.x && r.ur.x < ur.x &&
            ll.y < r.ll.y && r.ur.y < ur.y; }

 public:
  CsizeString<MG_RtreeStrSize_XyRange> cstr ();
  const char * str ();
};

/*
typedef XyRangeT<int>    XyRangeI;
typedef XyRangeT<double> XyRangeD;
*/

class XyRangeI : public XyRangeT<int> {
 public:
//  XyRangeI (XyPointI & ll, XyPointI & ur) : ll (ll), ur (ur) {}
  XyRangeI () {}
  XyRangeI (XyPointI & ll, XyPointI & ur) {
    this->ll = ll; this->ur = ur; }
 public:
  XyRange rngD ();
 public:
  CsizeString<MG_RtreeStrSize_XyRange> cstr ();
  const char * str ();
};

// MsJ tests
typedef struct HideTreeRange_ HideTreeRange;

class XyRangeD : public XyRangeT<double> {
 public:
  XyRangeD () {}
  XyRangeD (XyPointD & ll, XyPointD & ur) {
    this->ll = ll; this->ur = ur; }
  XyRangeD (XyPointI & ll, XyPointI & ur) {
    this->ll.x = ll.x; this->ll.y = ll.y;
    this->ur.x = ur.x; this->ur.y = ur.y; }
 public:
  XyRangeI rngI();
 public:
  void from (HideTreeRange & htr);
  XyRangeD (HideTreeRange & htr) { from (htr); }
 public:
  CsizeString<MG_RtreeStrSize_XyRange> cstr ();
  const char * str ();
};

inline void
XyRange::fromRngI (XyRangeI & r) { fromCoords (&r.ll.x); }

inline XyRangeI
XyRange::rngI () { return rngD().rngI(); }

//inline XyRangeD &
//XyRange::rngD () { return *((XyRangeD *)this); }

inline const char *
XyRange::str () { return rngD().str(); }

inline XyRangeI
XyRangeD::rngI() {
  XyRangeI r;
  r.ll.x = (int) floor(ll.x); r.ll.y = (int) floor(ll.y);
  r.ur.x = (int) ceil(ur.x);  r.ur.y = (int) ceil(ur.y);
  return r;
}

inline XyRange
XyRangeI::rngD() {
  XyRange r (*this); return r;
}


/*------------------------------------------------------------------------- o
o -------------------------------------------------------------------------*/

#endif      /* __MGRTREE_xyrangetH__ */






